#include "ep_lib.hpp"
#include <mpi.h>
#include "ep_declaration.hpp"
#include "ep_mpi.hpp"

namespace ep_lib
{

  int MPI_Comm_free(MPI_Comm *comm)
  {
    if(! (*comm)->is_ep)
    {
      return MPI_Comm_free_mpi(comm);
    }

    else 
    {
      if((*comm)->is_intercomm)
        return MPI_Comm_free_intercomm(comm);
      else
        return MPI_Comm_free_intracomm(comm);
    }
  }



  int MPI_Comm_free_mpi(MPI_Comm *comm)
  {
    Debug("MPI_Comm_free with MPI\n");

    return ::MPI_Comm_free(to_mpi_comm_ptr((*comm)->mpi_comm));
    
  }

  int MPI_Comm_free_intracomm(MPI_Comm *comm)
  {
    Debug("MPI_Comm_free with EP_intracomm\n");

    int ep_rank_loc, num_ep;

    ep_rank_loc = (*comm)->ep_comm_ptr->size_rank_info[1].first;
    num_ep = (*comm)->ep_comm_ptr->size_rank_info[1].second;

    MPI_Barrier(*comm);

    if(ep_rank_loc == 0)
    {

#ifdef _showinfo
      printf("delete (*comm)->my_buffer = %p\n", (*comm)->my_buffer);
#endif
      delete (*comm)->my_buffer;


      
#ifdef _showinfo        
      printf("delete (*comm)->ep_barrier = %p\n", (*comm)->ep_barrier);
#endif
      delete (*comm)->ep_barrier;
      


      (*comm)->ep_rank_map->clear();
#ifdef _showinfo
      printf("delete (*comm)->ep_rank_map = %p\n", (*comm)->ep_rank_map);
#endif
      delete (*comm)->ep_rank_map;
      

      for(int i=0; i<num_ep; i++)
      {
        (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr->message_queue->clear();
#ifdef _showinfo
        printf("delete (*comm)->ep_comm_ptr->comm_list[%d]->ep_comm_ptr->message_queue = %p\n", i, (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr->message_queue);
#endif
        delete (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr->message_queue;
          

#ifdef _showinfo
        printf("delete (*comm)->ep_comm_ptr->comm_list[%d]->ep_comm_ptr = %p\n", i, (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr);
#endif
        delete (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr;


#ifdef _showinfo
        printf("delete (*comm)->ep_comm_ptr->comm_list[%d] = %p\n", i, (*comm)->ep_comm_ptr->comm_list[i]);
#endif
        delete (*comm)->ep_comm_ptr->comm_list[i];
      }

#ifdef _showinfo
      printf("delete (*comm)->mpi_comm = %p\n", (*comm)->mpi_comm);
#endif
      ::MPI_Comm_free(to_mpi_comm_ptr((*comm)->mpi_comm));
      
#ifdef _showinfo
      printf("delete (*comm)->ep_comm_ptr->comm_list = %p\n", (*comm)->ep_comm_ptr->comm_list);
#endif
      delete[] (*comm)->ep_comm_ptr->comm_list;
    }
  }



  int MPI_Comm_free2(MPI_Comm *comm)
  {
    Debug("MPI_Comm_free with EP_intracomm\n");

    int ep_rank_loc, num_ep;

    ep_rank_loc = (*comm)->ep_comm_ptr->size_rank_info[1].first;
    num_ep = (*comm)->ep_comm_ptr->size_rank_info[1].second;

    MPI_Barrier2(*comm);

    if(ep_rank_loc == 0)
    {

#ifdef _showinfo
      printf("delete (*comm)->my_buffer = %p\n", (*comm)->my_buffer);
#endif
      delete (*comm)->my_buffer;


      
#ifdef _showinfo        
      printf("delete (*comm)->ep_barrier = %p\n", (*comm)->ep_barrier);
#endif
      delete (*comm)->ep_barrier;
      


      (*comm)->ep_rank_map->clear();
#ifdef _showinfo
      printf("delete (*comm)->ep_rank_map = %p\n", (*comm)->ep_rank_map);
#endif
      delete (*comm)->ep_rank_map;
      

      for(int i=0; i<num_ep; i++)
      {
        if((*comm)->is_intercomm)
        {
          (*comm)->ep_comm_ptr->comm_list[i]->inter_rank_map->clear();
#ifdef _showinfo
          printf("delete (*comm)->ep_comm_ptr->comm_list[%d]->inter_rank_map = %p\n", i, (*comm)->ep_comm_ptr->comm_list[i]->inter_rank_map);
#endif
          delete (*comm)->ep_comm_ptr->comm_list[i]->inter_rank_map;
        }

        (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr->message_queue->clear();
#ifdef _showinfo
        printf("delete (*comm)->ep_comm_ptr->comm_list[%d]->ep_comm_ptr->message_queue = %p\n", i, (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr->message_queue);
#endif
        delete (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr->message_queue;
          

#ifdef _showinfo
        printf("delete (*comm)->ep_comm_ptr->comm_list[%d]->ep_comm_ptr = %p\n", i, (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr);
#endif
        delete (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr;


#ifdef _showinfo
        printf("delete (*comm)->ep_comm_ptr->comm_list[%d] = %p\n", i, (*comm)->ep_comm_ptr->comm_list[i]);
#endif
        delete (*comm)->ep_comm_ptr->comm_list[i];
      }

#ifdef _showinfo
      printf("delete (*comm)->mpi_comm = %p\n", (*comm)->mpi_comm);
#endif
      ::MPI_Comm_free(to_mpi_comm_ptr((*comm)->mpi_comm));
      
#ifdef _showinfo
      printf("delete (*comm)->ep_comm_ptr->comm_list = %p\n", (*comm)->ep_comm_ptr->comm_list);
#endif
      delete[] (*comm)->ep_comm_ptr->comm_list;
    }
  }



  int MPI_Comm_free_intercomm(MPI_Comm *comm)
  {
    int ep_rank;
    MPI_Comm_rank(*comm, &ep_rank);
    int ep_rank_loc = (*comm)->ep_comm_ptr->size_rank_info[1].first;
    int num_ep =      (*comm)->ep_comm_ptr->size_rank_info[1].second;
    
    int newcomm_ep_rank =(*comm)->ep_comm_ptr->intercomm->size_rank_info[0].first; 
    int newcomm_ep_rank_loc = (*comm)->ep_comm_ptr->intercomm->size_rank_info[1].first;
    int newcomm_num_ep = (*comm)->ep_comm_ptr->intercomm->size_rank_info[1].second;

    MPI_Barrier(*comm);

    if(ep_rank_loc == 0)
    {
      (*comm)->ep_comm_ptr->intercomm->intercomm_rank_map->clear();
#ifdef _showinfo
      printf("delete (*comm)->ep_comm_ptr->intercomm->intercomm_rank_map = %p\n", (*comm)->ep_comm_ptr->intercomm->intercomm_rank_map);
#endif
      delete (*comm)->ep_comm_ptr->intercomm->intercomm_rank_map;

      (*comm)->ep_comm_ptr->intercomm->local_rank_map->clear();
#ifdef _showinfo
      printf("delete (*comm)->ep_comm_ptr->intercomm->local_rank_map = %p\n", (*comm)->ep_comm_ptr->intercomm->local_rank_map);
#endif
      delete (*comm)->ep_comm_ptr->intercomm->local_rank_map;
    }

    if(newcomm_ep_rank_loc == 0)
    {

#ifdef _showinfo
      printf("delete (*comm)->my_buffer = %p\n", (*comm)->my_buffer);
#endif
      delete (*comm)->my_buffer;


      
#ifdef _showinfo        
      printf("delete (*comm)->ep_barrier = %p\n", (*comm)->ep_barrier);
#endif
      delete (*comm)->ep_barrier;
      

      (*comm)->ep_rank_map->clear();
#ifdef _showinfo
      printf("delete (*comm)->ep_rank_map = %p\n", (*comm)->ep_rank_map);
#endif
      delete (*comm)->ep_rank_map;
      
#ifdef _showinfo
      printf("delete (*comm)->ep_comm_ptr->intercomm->mpi_inter_comm = %p\n", (*comm)->ep_comm_ptr->intercomm->mpi_inter_comm);
#endif
      ::MPI_Comm_free(to_mpi_comm_ptr((*comm)->ep_comm_ptr->intercomm->mpi_inter_comm));

      for(int i=0; i<newcomm_num_ep; i++)
      {
        (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr->message_queue->clear();
#ifdef _showinfo
        printf("delete (*comm)->ep_comm_ptr->comm_list[%d]->ep_comm_ptr->message_queue = %p\n", i, (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr->message_queue);
#endif
        delete (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr->message_queue;

#ifdef _showinfo
        printf("delete (*comm)->ep_comm_ptr->comm_list[%d]->ep_comm_ptr->intercomm = %p\n", i, (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr->intercomm);
#endif
        delete (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr->intercomm;        
          

#ifdef _showinfo
        printf("delete (*comm)->ep_comm_ptr->comm_list[%d]->ep_comm_ptr = %p\n", i, (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr);
#endif
        delete (*comm)->ep_comm_ptr->comm_list[i]->ep_comm_ptr;


#ifdef _showinfo
        printf("delete (*comm)->ep_comm_ptr->comm_list[%d] = %p\n", i, (*comm)->ep_comm_ptr->comm_list[i]);
#endif
        delete (*comm)->ep_comm_ptr->comm_list[i];

      }

#ifdef _showinfo
      printf("delete (*comm)->mpi_comm = %p\n", (*comm)->mpi_comm);
#endif
      ::MPI_Comm_free(to_mpi_comm_ptr((*comm)->mpi_comm));
      
#ifdef _showinfo
      printf("delete (*comm)->ep_comm_ptr->comm_list = %p\n", (*comm)->ep_comm_ptr->comm_list);
#endif
      delete[] (*comm)->ep_comm_ptr->comm_list;
    }
  }


}


